import pytest
import tempfile
import os


from az.cache import FileCache

@pytest.fixture
def file_cache():
    temp_dir = tempfile.mkdtemp()
    cache_file = os.path.join(temp_dir, 'test_cache.json')
    cache = FileCache(cache_file)
    yield cache
    os.remove(cache_file)
    os.rmdir(temp_dir)

def test_set_and_get(file_cache):
    file_cache.set('test_key', ['value1', 'value2'])
    assert file_cache.get('test_key') == {'value1', 'value2'}

def test_update(file_cache):
    file_cache.set('test_key', ['value1'])
    file_cache.update('test_key', ['value2', 'value3'])
    assert file_cache.get('test_key') == {'value1', 'value2', 'value3'}

def test_clear(file_cache):
    file_cache.set('test_key', ['value1'])
    file_cache.clear()
    assert file_cache.get('test_key') == set()

def test_persistence(file_cache):
    file_cache.set('test_key', ['value1'])
    cache_file = file_cache.cache_file
    del file_cache
    new_cache = FileCache(cache_file)
    assert new_cache.get('test_key') == {'value1'}


import os


from az.llm_provider import LLMProvider
from openai import OpenAI, NotFoundError, RateLimitError, AuthenticationError, APIError
from az.cache import FileCache



MODELS_CACHE_FILE = os.path.expanduser("~/.config/.azc_models.json" if os.path.exists(os.path.expanduser("~/.config")) else "~/.azc_models.json")

class OpenAIClient(LLMProvider):
    def __init__(self, config={}, primer=None):
        self.provider = 'openai'
        self.client = OpenAI()
        self.models_cache = FileCache(MODELS_CACHE_FILE)
        
        self.list_models()
        self.config = config
        self.model = self.config.get("openai", {}).get("model", "gpt-4o-mini")
        self.messages = []
        
        self.primer = primer
        if self.primer:
            self.messages.append({"role": "system", "content": self.primer})
          

    def list_models(self):
        self.models = self.models_cache.get(self.provider)
        if len(self.models) == 0:
            self.refresh_models()
        return self.models


    def refresh_models(self):
        print("refreshing models")
        try:
            models = [m.id for m in self.client.models.list().data]
            self.models = models
            self.models_cache.set(self.provider, models)
            print(f"Got {len(self.models)} models. Type 'l' to list them.")
        except NotFoundError as e:
            # sometimes I get a 404 on this
            print(f"Error fetching models: {e}. Either use a different model or restart and try again.")


    def chat(self, message):
        if self.primer and len(self.messages) == 0:
            self.messages.append({"role": "user", "content": self.primer + "\n\n" + message})
        else:
            self.messages.append({"role": "user", "content": message})
        
        try:
            response_stream = self.client.chat.completions.create(
                model=self.model,
                messages=self.messages,
                stream=True,
            )
            
            collected_messages = []
            for chunk in response_stream:
                if chunk.choices[0].delta.content is not None:
                    collected_messages.append(chunk.choices[0].delta.content)
                    yield chunk.choices[0].delta.content

            full_response = ''.join(collected_messages)
            self.messages.append({"role": "assistant", "content": full_response})
            
        except RateLimitError as e:
            error_message = "OpenAI API rate limit exceeded or credits exhausted. Please check your usage and limits."
            print(f"\n[red]{error_message}[/red]")
            return
        except AuthenticationError as e:
            error_message = "OpenAI API authentication failed. Please check your API key."
            print(f"\n[red]{error_message}[/red]")
            return
        except APIError as e:
            error_message = f"OpenAI API error: {str(e)}"
            print(f"\n[red]{error_message}[/red]")
            return
        except Exception as e:
            error_message = f"Unexpected error: {str(e)}"
            print(f"\n[red]{error_message}[/red]")
            return


if __name__ == "__main__": # pragma: no cover
    client = OpenAIClient(primer="Limit your response to 300 characters or less")
    for text in client.chat("I'm traveling to Madrid soon (mid-October) with my wife. We love food, history and shopping. We've been there before. Can you recommend a few destinations/activities off the beaten path? We're staying in the city center and will be there for 3 days. We're looking for authentic experiences, not tourist traps."):
        print(text, end="", flush=True)
    print()

    print(client.list_models())



import json
import os

import requests
import httpx

from az.llm_provider import LLMProvider


class OllamaClient(LLMProvider):
    def __init__(self, config={}):
        LLMProvider.__init__(self)
        self.provider = 'ollama'
        self.name = 'ollama'
        base_url = os.getenv("OLLAMA_URL", "http://localhost:11434")
        base_url = base_url.rstrip('/')
        self.client = httpx.Client(
            base_url=base_url,
            timeout=40.0  # Set timeout to 40 seconds
        )
        self.config = config
        self.models = self.list_models()
        # Get model from config or use first available model as default
        self.model = self.config.get("ollama", {}).get("model") or (self.models[0] if self.models else "llama2")
        self.messages = []
        self.primer = config.get("primer")
        if self.primer:
            self.messages.append({"role": "system", "content": self.primer})

    def list_models(self):
        """List all models available from Ollama's /api/tags endpoint"""
        try:
            response = self.client.get("/api/tags")
            response.raise_for_status()
            
            if response.status_code == 200:
                models_data = response.json()["models"]
                return [model["name"] for model in models_data]
            else:
                print(f"Warning: Could not list models, using empty list")
                return []
                
        except Exception as e:
            print(f"Warning: Could not list models: {e}")
            return []

    @property
    def model(self):
        return self._model

    @model.setter
    def model(self, value):
        # For Ollama, just accept any model name as it might be pulled on demand
        self._model = value

    def chat(self, message):
        if self.primer and len(self.messages) == 0:
            self.messages.append({"role": "user", "content": self.primer + "\n\n" + message})
        else:
            self.messages.append({"role": "user", "content": message})

        try:
            with self.client.stream('POST', "/api/chat", json={
                "model": self.model,
                "messages": self.messages,
                "stream": True
            }) as response:
                response.raise_for_status()
                collected_message = []

                for chunk in response.iter_bytes():
                    if not chunk:
                        continue
                        
                    try:
                        data = json.loads(chunk)
                        if data.get('done', False):
                            continue
                        if 'message' in data and 'content' in data['message']:
                            content = data['message']['content']
                            # Skip the thinking tokens if present.
                            if content == '<think>':
                                content = 'thinking...\n'
                            elif content == '</think>':
                                content = '... done thinking.\n'
                            collected_message.append(content)
                            yield content
                    except json.JSONDecodeError:
                        continue

                # Add the complete message to history
                full_response = ''.join(collected_message)
                self.messages.append({"role": "assistant", "content": full_response})

        except httpx.HTTPError as e:
            error_message = f"Ollama API error: {str(e)}"
            print(f"\n[red]{error_message}[/red]")
            return
        except Exception as e:
            error_message = f"Unexpected error: {str(e)}"
            print(f"\n[red]{error_message}[/red]")
            return


def test_ollama(model="llama2", prompt="Tell me a short story about a robot"):
    """Test function for Ollama provider"""
    print(f"\nTesting Ollama with model: {model}")
    print(f"Prompt: {prompt}")
    print("\nResponse:")
    
    try:
        client = OllamaClient({"ollama": {"model": model}})
        print("Models:")
        print(client.list_models())
        for chunk in client.chat(prompt):
            print(chunk, end='', flush=True)
        print("\n\nTest completed.")
        
    except Exception as e:
        print(f"\nError during test: {str(e)}")

if __name__ == "__main__":
    import sys
    
    # Get model and prompt from command line arguments, or use defaults
    model = sys.argv[1] if len(sys.argv) > 1 else "llama2"
    prompt = sys.argv[2] if len(sys.argv) > 2 else "Tell me a short story about a robot"
    
    test_ollama(model, prompt)



from az.openai_provider import OpenAIClient
from az.llm_provider import LLMProvider
import os
from openai import OpenAI

class GrokClient(OpenAIClient):
    def __init__(self, config={}, primer=None):
        LLMProvider.__init__(self, primer)  # Call this first
        self.provider = 'grok'
        self.name = 'grok'
        self.client = OpenAI(
            api_key=os.getenv("XAI_API_KEY"),
            base_url="https://api.x.ai/v1"
        )
        self.config = config
        self.models = self.list_models()
        self.model = self.config.get("grok", {}).get("model", "grok-beta")
        
    def list_models(self):
        """Grok currently only has one model"""
        return ["grok-beta"]

if __name__ == "__main__": # pragma: no cover
    client = GrokClient(primer="You are Grok, inspired by Hitchhiker's Guide")
    for text in client.chat("What is the meaning of life?"):
        print(text, end="", flush=True)
    print() 


import os
import json
import importlib.resources
import pkg_resources

_config = None  # Cache for config
_config_path = None  # Store the path we loaded from

DEFAULT_CONFIG = {
    "default-provider": "openai",
    "default-model": "gpt-4",
    "default-models": {
        "openai": {"model": "gpt-4"},
        "anthropic": {"model": "claude-3-opus-20240229"},
        "gemini": {"model": "gemini-pro"},
        "ollama": {"model": "llama2"},
        "grok": {"model": "grok-1"}
    }
}

def load_config(config_path=None):
    """Load configuration with precedence
    Args:
        config_path: Optional path to config file (used for testing)
    """
    global _config, _config_path
    if _config is not None and config_path is None:
        return _config

    # If a specific path is provided (for testing), try to use it
    if config_path is not None:
        try:
            with open(config_path, 'r') as f:
                return json.load(f)
        except FileNotFoundError:
            # For testing, return empty dict when file not found
            return {}

    # Try current directory first
    local_config = os.path.join(os.getcwd(), 'config.json')
    if os.path.exists(local_config):
        _config_path = os.path.abspath(local_config)
        with open(local_config, 'r') as f:
            _config = json.load(f)
            print(f"Config: {_config_path}")
            return _config

    # Try user config directory
    user_config_dir = os.path.expanduser("~/.config/azc")
    user_config = os.path.join(user_config_dir, 'config.json')
    if os.path.exists(user_config):
        _config_path = os.path.abspath(user_config)
        with open(user_config, 'r') as f:
            _config = json.load(f)
            print(f"Config: {_config_path}")
            return _config

    # Use project's default config
    default_config = os.path.join(os.path.dirname(os.path.dirname(__file__)), 'config.json')
    if os.path.exists(default_config):
        _config_path = os.path.abspath(default_config)
        with open(default_config, 'r') as f:
            _config = json.load(f)
            print(f"Config: using defaults from {_config_path}")
            print(f"To customize: create config.json in current dir or ~/.config/azc/")
            return _config
            
    # If no config files found, return empty dict
    return {}

def default_model(config_path=None):
    """Get the default model from config"""
    config = load_config(config_path)
    if "default-models" in config:
        provider = config.get("default-provider")
        if provider and provider in config["default-models"]:
            return config["default-models"][provider].get("model")
    return None

def default_provider(config_path=None):
    """Get the default provider from config"""
    config = load_config(config_path)
    # For non-existent file, return "openai" as default
    if config_path and not os.path.exists(config_path):
        return "openai"
    return config.get("default-provider")

if __name__ == "__main__":
    config = load_config()
    print(json.dumps(config, indent=4))

    print("default provider:", default_provider())
    print("default model:", default_model())

    config_file = "does-not-exist.json"
    config = load_config(config_file)
    print(json.dumps(config, indent=4))

    print("default provider:", default_provider(config_file))
    print("default model:", default_model(config_file))
    



import pytest
from prompt_toolkit.application import create_app_session
from prompt_toolkit.input import create_pipe_input
from prompt_toolkit.output import DummyOutput
from prompt_toolkit import PromptSession
from sample_app import main
from unittest.mock import patch
from io import StringIO

@pytest.fixture(autouse=True, scope="function")
def mock_input():
    with create_pipe_input() as pipe_input:
        with create_app_session(input=pipe_input, output=DummyOutput()):
            yield pipe_input


def test_sample(mock_input):
    mock_input.send_text("test\n\nq\n")
    session = PromptSession()
    result = session.prompt("")
    assert result == "test"


def test_sample_app(mock_input):
    mock_input.send_text("4\n")
    out = StringIO()
    with patch('sys.stdout', new=out) as fake_out:
        main()
    assert out.getvalue().strip() == "4"


import pytest
from az.config import load_config, default_provider, default_model

@pytest.fixture
def config_file():
    return "config.json"

def test_load_config(config_file, tmp_path):
    # Create a temporary config file
    temp_config = tmp_path / config_file
    temp_config.write_text('{"provider": "openai", "model": "gpt-3.5-turbo"}')
    
    config = load_config(str(temp_config))
    assert isinstance(config, dict)
    assert config["provider"] == "openai"
    assert config["model"] == "gpt-3.5-turbo"

def test_load_config_file_not_found(config_file):
    config = load_config("non_existent_config.json")
    assert isinstance(config, dict)
    assert len(config) == 0

def test_default_provider(tmp_path):
    # Create a temporary config file
    temp_config = tmp_path / "test_config.json"
    temp_config.write_text('{"default-provider": "ollama"}')
    
    provider = default_provider(str(temp_config))
    assert isinstance(provider, str)
    assert provider == "ollama"

def test_default_provider_non_existent_file(tmp_path):
    # Test with non-existent file
    non_existent_config = tmp_path / "non_existent.json"
    default_prov = default_provider(str(non_existent_config))
    assert default_prov == "openai"  # Assuming "openai" is the fallback default

def test_default_model(tmp_path):
    # Create a temporary config file
    temp_config = tmp_path / "test_config.json"
    print(f"temp_config: {temp_config}")
    temp_config.write_text('{"default-provider": "openai", "default-models": {"openai": {"model":"gpt-4"}}}')
    
    model = default_model(str(temp_config))
    assert model == "gpt-4"

def test_default_model_non_existent_file(tmp_path):
    # Test with non-existent file
    non_existent_config = tmp_path / "non_existent.json"
    default_mod = default_model(str(non_existent_config))
    assert default_mod is None  # Assuming None is returned when no config file exists

def test_load_config_with_missing_fields(config_file, tmp_path):
    # Create a temporary config file with missing fields
    temp_config = tmp_path / config_file
    temp_config.write_text('{"provider": "openai"}')
    
    config = load_config(str(temp_config))
    assert isinstance(config, dict)
    assert config["provider"] == "openai"
    assert default_model(str(temp_config)) == None, f"default_model should return None if model is not specified in file {str(temp_config)}"


def test_load_config_with_invalid_json(config_file, tmp_path):
    # Create a temporary config file with invalid JSON
    temp_config = tmp_path / config_file
    temp_config.write_text('{"provider": "openai", "model":}')
    
    with pytest.raises(ValueError):  # Assuming the function raises a ValueError for invalid JSON
        load_config(str(temp_config))




from az.ollama_provider import OllamaClient

import pytest
from az.ollama_provider import OllamaClient
from az.openai_provider import OpenAIClient
from az.anthropic_provider import AnthropicClient
from az.gemini_provider import GeminiClient

@pytest.fixture(params=[OllamaClient, OpenAIClient, AnthropicClient, GeminiClient])
def client(request):
    return request.param(primer="Limit your response to 300 characters or less")

@pytest.mark.expensive
@pytest.mark.parametrize("question,expected", [
    # ("how much is 6^2?", "36"),
    ("what is the capital of France?", "Paris"),
    # ("who wrote 'To Kill a Mockingbird'?", "Harper Lee"),
])
def test_providers(client, question, expected):
    response = ""
    for text in client.chat(question):
        response += text
    assert expected in response
    assert client.n_user_messages() == 1
    client.new_chat()
    assert client.n_user_messages() == 0





# chat_app.py

import sys
import os
import shutil
import argparse
import importlib

# a mix of rich and prompt_toolkit seem to hit the sweet spot for terminal UI interactivity
from rich.console import Console
from rich.markdown import Markdown
from rich.panel import Panel
from rich.align import Align
from rich.live import Live
from rich.box import Box
from prompt_toolkit import PromptSession
from prompt_toolkit.patch_stdout import patch_stdout
from prompt_toolkit.formatted_text import HTML
from prompt_toolkit.history import FileHistory
from prompt_toolkit.completion import Completer, Completion
from prompt_toolkit.key_binding import KeyBindings


import readline # needed for prompt editing

from az.utils import number_to_ordinal
from az.config import load_config, default_model, default_provider

HISTORY_FILE_NAME = os.path.expanduser("~/.config/.azc_history" if os.path.exists(os.path.expanduser("~/.config")) else "~/.azc_history")

config = load_config()

bindings = KeyBindings()

def is_command(string: str) -> bool:
    if string.strip().lower() in ('exit', 'quit', 'q', 'p', 'l', 'm', 'p', 'n', 'h', 'r', '?', ''): return True
    elif string.strip().startswith("p "): return True
    else: return False


def get_input():
    # added so we can override input for testing
    return None


# Bind 'Control+n' to insert a newline
@bindings.add('c-n')
def insert_newline(event):
    event.current_buffer.insert_text('\n')


# An empty box border makes it easier to copy and paste.
EMPTY: Box = Box(
    "    \n"
    "    \n"
    "    \n"
    "    \n"
    "    \n"
    "    \n"
    "    \n"
    "    \n",
    ascii=True,
)

# Initialize the console
console = Console()

# this is done by the providers so it's not needed here
# load_dotenv(os.path.expanduser("~/.config/.env" if os .path.exists(os.path.expanduser("~/.config")) else "~/.env"))


providers = []
def discover_providers():
    """Discover providers based on config file"""
    available_providers = {}
    provider_configs = config.get("default-models", {}) if config else {}
    
    if not provider_configs:
        print("Warning: No provider configurations found in config")
        return available_providers
    
    for provider_name in provider_configs:
        try:
            module_name = f"az.{provider_name}_provider"
            class_name = "OpenAIClient" if provider_name == "openai" else f"{provider_name.capitalize()}Client"
            
            module = importlib.import_module(module_name)
            client_class = getattr(module, class_name)
            available_providers[provider_name] = client_class
            
            # Check for environment variables
            if provider_name == "grok" and "XAI_API_KEY" in os.environ:
                providers.append(provider_name)
            elif provider_name == "ollama" and "OLLAMA_URL" in os.environ:
                providers.append(provider_name)
            elif f"{provider_name.upper()}_API_KEY" in os.environ:
                providers.append(provider_name)
                
        except Exception as e:
            print(f"Warning: Could not load provider {provider_name}: {e}")
    
    if not providers:
        # Make the error message more helpful
        print("No provider API keys found in environment.")
        print("Please set one or more of the following in your .env file:")
        for provider in available_providers.keys():
            if provider == "grok":
                print(f"  XAI_API_KEY for {provider}")
            elif provider == "ollama":
                print(f"  OLLAMA_URL for {provider}")
            else:
                print(f"  {provider.upper()}_API_KEY for {provider}")
    
    return available_providers

PROVIDERS = discover_providers()

# provider_completer = WordCompleter([f"p {provider}" for provider in providers], ignore_case=True)

class CommandsCompleter(Completer):
    """
    This completer handles commands and model names
    """
    def get_completions(self, document, complete_event):
        text = document.current_line
        
        # Provider completion (existing functionality)
        if text.startswith('p '):   
            for provider in providers:
                yield Completion(
                    f'p {provider}', 
                    start_position=-len(text),
                    display=HTML(f'{provider}'),
                    style='bg:ansiyellow'
                )
                
        # Model completion (new functionality)
        elif text.startswith('m '):
            prefix = text[len('m '):].lower()
            if client and hasattr(client, 'models'):
                for model in client.models:
                    if model.lower().startswith(prefix):
                        yield Completion(
                            model,
                            start_position=-len(prefix),
                            display=HTML(f'{model}'),
                            style='bg:ansiblue'
                        )

completer = CommandsCompleter()


def primer():
    _, rows = shutil.get_terminal_size()
    # single screen is better...
    return f"please limit your response to {rows-4} lines at most)"


def provider_factory(provider_hint):
    provider_found = False
    provider_full_name = None

    for provider_name in PROVIDERS.keys():
        if provider_hint in provider_name:
            provider_full_name = provider_name
            provider_found = True
            break

    if not provider_found:
        raise ValueError(f"Cannot find provider with <{provider_hint}>")
    
    # Get the default model for this provider from config
    provider_config = config.get("default-models", {}).get(provider_full_name, {})
    configured_model = provider_config.get("model")
    
    client = PROVIDERS[provider_full_name]({
        **config,
        "primer": primer(),
    })
    
    # Set the model if specified in config, ensuring exact match
    if configured_model:
        if configured_model not in client.models:
            console.print(f"[yellow]Warning: Configured model '{configured_model}' not found in available models for {provider_full_name}[/]")
            console.print(f"Available models:")
            for model in client.models:
                console.print(f"  - {model}")
        else:
            client.model = configured_model
            
    return client



class FilteredHistory(FileHistory):
    """
    This class is a custom history class that filters out commands we don't want to save
    """
    def store_string(self, string: str) -> None:
        if not is_command(string):
                super().store_string(string)


def help():
    return """

Just type your message and press enter to start a chat.

| Command | Description |
|---------|-------------|
| l       | List models for current provider |
| r       | Refresh models for current provider |
| n       | New chat (forget history)   |
| ? or h  | Help (this screen) |
| m       | Change model for current provider |
| p provider_name | Change provider (p and space trigger autocomplete) |
| ctrl-n  | New line |
"""



def main(initial_prompt=None):
    """Main entry point"""
    global client, config, PROVIDERS
    
    if len(providers) == 0:
        console.print('no providers found, exiting, please set one of the following: OPENAI_API_KEY, OLLAMA_URL, ANTHROPIC_API_KEY, GEMINI_API_KEY in a .env file')
        return
    
    parser = argparse.ArgumentParser(description="Chat with an AI assistant")
    parser.add_argument("-p", "--provider", help="The provider to use, e.g. 'openai' or 'ollama'. Abbreviations allowed, like 'op' for 'openai'")
    parser.add_argument("-m", "--model", help="The model to use")
    parser.add_argument("-d", "--double-enter", action="store_true", help="Enable 'press enter twice to submit' mode")
    parser.add_argument("-b", "--batch", action="store_true", help="Not interactive, just do one chat and exit")
    parser.add_argument("-v", "--verbose", action="store_true", help="Verbose output")
    parser.add_argument("initial_prompt", nargs='?', help="Initial prompt (if provided without a flag)")
    args = parser.parse_args()
    initial_prompt = args.initial_prompt

    if args.verbose:
        console.print('providers configured: [yellow]' + ', '.join(providers) + '[/]')

    if args.double_enter:
        console.print("type <enter> twice to submit...")
        @bindings.add('enter')
        def _(event):
            buffer = event.current_buffer
            accepted = False
            if buffer.document.text.strip() == '':
                accepted = True
            elif is_command(buffer.document.text):
                accepted = True
            elif buffer.document.text.endswith('\n') and buffer.document.is_cursor_at_the_end:
                accepted = True
            else:
                accepted = False
            if accepted:
                buffer.validate_and_handle()
            else:
                buffer.insert_text('\n')



    provider_name = args.provider if args.provider else default_provider()
    provider_name = provider_name if provider_name else providers[0]
    
    client = provider_factory(provider_name)

    if args.model:
        client.model = args.model
    
    if args.verbose:
        console.print(f'using: [green]{client}[/]')
        console.print('[magenta]type ? or h for help[/]')

    def bottom_toolbar():
        return HTML(f' Using <b>{client}</b> ({number_to_ordinal(client.n_user_messages()+1)} message)     <ansicyan>enter ? or h for help</ansicyan> ')

    our_history = FilteredHistory(HISTORY_FILE_NAME)
    session = PromptSession(history=our_history, input=get_input())

    done=False
    
    try:
        while not done:
            if initial_prompt:
                user_input = initial_prompt
                initial_prompt = None
                if args.batch:
                    done=True
            else:
                with patch_stdout():
                    try:
                        user_input = session.prompt(
                            HTML(f'<ansicyan>azc></ansicyan> '),
                            completer=completer,
                            bottom_toolbar=bottom_toolbar,
                            key_bindings=bindings
                        )
                    except EOFError:
                        done = True
                        break


            if user_input.strip().lower() == '':
                # Some people like to press enter to get a new line
                continue

            if user_input.strip().lower() in ('exit', 'quit', 'q'):
                break

            if user_input.strip().lower() in ('l'):
                console.print(Markdown('  - ' + '\n  - '.join(client.list_models())))
                continue

            if user_input.strip().lower() in ('n'):
                console.print(f'new chat with {client}')
                client.new_chat()
                continue

            if user_input.strip().lower() in ('r'):
                client.refresh_models()
                continue

            if user_input.strip().lower() in ('h', '?'):
                console.print(Markdown(help()))
                continue

            if user_input.startswith('m '):
                model_name = user_input.split(' ')[1]
                if model_name:
                    try:
                        client.model = model_name
                        console.print(f'using: {client.provider}:{client.model}')
                    except ValueError as e:
                        console.print(f"[red]Error: {str(e)}[/]")
                    continue
                else:
                    console.print("Model selection cancelled")
            

            if user_input.startswith('p '):
                provider_name = user_input.split(' ')[1]
                try:
                    client = provider_factory(provider_name)
                    console.print(f'using: {client} (new chat)')
                except ValueError as e:
                    console.print(f'[red]error: {e}[/]')
                continue

            title = f"{client} ({number_to_ordinal(client.n_user_messages()+1)} message)" if not args.batch else None

            assistant_panel = Panel(
                Align.left(""),
                title=title,
                style="yellow",
                expand=True,
                box=EMPTY
            )

            current_message = ""

            # Note that vertical_overflow="visible" causes realtime updates of rendered markdown beyond the full window height,
            # but it leaves a trail of partially rendered markdown behind when new content is added, hence it is not used
            with Live(assistant_panel, console=console, refresh_per_second=2, vertical_overflow='ellipsis') as live:
                if args.double_enter:
                    console.print(f'...')
                for chunk in client.chat(user_input):
                    current_message += chunk
                    new_panel = Panel(
                        Align.left(Markdown(current_message)),
                        title=title,
                        style="yellow",
                        expand=True,
                        border_style="none",
                        box=EMPTY
                    )
                    live.update(new_panel, refresh=False)

    except KeyboardInterrupt:
        done = True
    finally:
        if not args.batch:
            console.print(":wave: [italic]Bye[/]")

if __name__ == "__main__": # pragma: no cover
    main()


